from __future__ import annotations

import gzip
import platform
import subprocess
import pipes
import shutil
import os
import tempfile
from typing import Callable

import dangerzone.util as dzutil


def exec_container(args, stdout_callback: Callable[[str], None]) -> int:
    args = [dzutil.CONTAINER_RUNTIME] + args
    args_str = " ".join(pipes.quote(arg) for arg in args)
    print("> " + args_str)

    with subprocess.Popen(
        args,
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True,
        startupinfo=dzutil.get_subprocess_startupinfo(),
    ) as p:
        if p.stdout:
            for line in p.stdout:
                stdout_callback(line)
        p.communicate()
        return p.returncode


def convert(
    input_filename, output_filename, ocr_lang, stdout_callback: Callable[[str], None]
):
    success = False

    if ocr_lang:
        ocr = "1"
    else:
        ocr = "0"

    dz_tmp = os.path.join(dzutil.APPDATA_PATH, "tmp")
    os.makedirs(dz_tmp, exist_ok=True)

    tmpdir = tempfile.TemporaryDirectory(dir=dz_tmp)
    pixel_dir = os.path.join(tmpdir.name, "pixels")
    safe_dir = os.path.join(tmpdir.name, "safe")
    os.makedirs(pixel_dir, exist_ok=True)
    os.makedirs(safe_dir, exist_ok=True)

    platform_args: list[str] = []
    if dzutil.CONTAINER_COMMAND == "docker":
        platform_args += ["--platform", "linux/amd64"]
    else:
        platform_args = []

    # Convert document to pixels
    args = (
        ["run", "--network", "none"]
        + platform_args
        + [
            "-v",
            f"{input_filename}:/tmp/input_file",
            "-v",
            f"{pixel_dir}:/dangerzone",
            dzutil.CONTAINER_NAME,
            "/usr/bin/python3",
            "/usr/local/bin/dangerzone.py",
            "document-to-pixels",
        ]
    )
    ret = exec_container(args, stdout_callback)
    if ret != 0:
        print("documents-to-pixels failed")
    else:
        # TODO: validate convert to pixels output

        # Convert pixels to safe PDF
        args = (
            ["run", "--network", "none"]
            + platform_args
            + [
                "-v",
                f"{pixel_dir}:/dangerzone",
                "-v",
                f"{safe_dir}:/safezone",
                "-e",
                f"OCR={ocr}",
                "-e",
                f"OCR_LANGUAGE={ocr_lang}",
                dzutil.CONTAINER_NAME,
                "/usr/bin/python3",
                "/usr/local/bin/dangerzone.py",
                "pixels-to-pdf",
            ]
        )
        ret = exec_container(args, stdout_callback)
        if ret != 0:
            print("pixels-to-pdf failed")
        else:
            # Move the final file to the right place
            if os.path.exists(output_filename):
                os.remove(output_filename)

            container_output_filename = os.path.join(
                safe_dir, "safe-output-compressed.pdf"
            )
            shutil.move(container_output_filename, output_filename)

            # We did it
            success = True

    # Clean up
    tmpdir.cleanup()

    return success


# From global_common:

# def validate_convert_to_pixel_output(self, common, output):
#     """
#     Take the output from the convert to pixels tasks and validate it. Returns
#     a tuple like: (success (boolean), error_message (str))
#     """
#     max_image_width = 10000
#     max_image_height = 10000

#     # Did we hit an error?
#     for line in output.split("\n"):
#         if (
#             "failed:" in line
#             or "The document format is not supported" in line
#             or "Error" in line
#         ):
#             return False, output

#     # How many pages was that?
#     num_pages = None
#     for line in output.split("\n"):
#         if line.startswith("Document has "):
#             num_pages = line.split(" ")[2]
#             break
#     if not num_pages or not num_pages.isdigit() or int(num_pages) <= 0:
#         return False, "Invalid number of pages returned"
#     num_pages = int(num_pages)

#     # Make sure we have the files we expect
#     expected_filenames = []
#     for i in range(1, num_pages + 1):
#         expected_filenames += [
#             f"page-{i}.rgb",
#             f"page-{i}.width",
#             f"page-{i}.height",
#         ]
#     expected_filenames.sort()
#     actual_filenames = os.listdir(common.pixel_dir.name)
#     actual_filenames.sort()

#     if expected_filenames != actual_filenames:
#         return (
#             False,
#             f"We expected these files:\n{expected_filenames}\n\nBut we got these files:\n{actual_filenames}",
#         )

#     # Make sure the files are the correct sizes
#     for i in range(1, num_pages + 1):
#         with open(f"{common.pixel_dir.name}/page-{i}.width") as f:
#             w_str = f.read().strip()
#         with open(f"{common.pixel_dir.name}/page-{i}.height") as f:
#             h_str = f.read().strip()
#         w = int(w_str)
#         h = int(h_str)
#         if (
#             not w_str.isdigit()
#             or not h_str.isdigit()
#             or w <= 0
#             or w > max_image_width
#             or h <= 0
#             or h > max_image_height
#         ):
#             return False, f"Page {i} has invalid geometry"

#         # Make sure the RGB file is the correct size
#         if os.path.getsize(f"{common.pixel_dir.name}/page-{i}.rgb") != w * h * 3:
#             return False, f"Page {i} has an invalid RGB file size"

#     return True, True
def is_container_installed():
    """
    See if the podman container is installed. Linux only.
    """
    # Get the image id
    with open(dzutil.get_resource_path("image-id.txt")) as f:
        expected_image_id = f.read().strip()

    # See if this image is already installed
    installed = False
    found_image_id = subprocess.check_output(
        [
            dzutil.CONTAINER_RUNTIME,
            "image",
            "list",
            "--format",
            "{{.ID}}",
            dzutil.CONTAINER_NAME,
        ],
        text=True,
        startupinfo=dzutil.get_subprocess_startupinfo(),
    )
    found_image_id = found_image_id.strip()

    if found_image_id == expected_image_id:
        installed = True
    elif found_image_id == "":
        pass
    else:
        print("Deleting old dangerzone container image")

        try:
            subprocess.check_output(
                [dzutil.CONTAINER_RUNTIME, "rmi", "--force", found_image_id],
                startupinfo=dzutil.get_subprocess_startupinfo(),
            )
        except:
            print("Couldn't delete old container image, so leaving it there")

    return installed


def install_container():
    """
    Make sure the podman container is installed. Linux only.
    """
    if is_container_installed():
        return

    # Load the container into podman
    print("Installing Dangerzone container image...")

    p = subprocess.Popen(
        [dzutil.CONTAINER_RUNTIME, "load"],
        stdin=subprocess.PIPE,
        startupinfo=dzutil.get_subprocess_startupinfo(),
    )

    chunk_size = 10240
    compressed_container_path = dzutil.get_resource_path("container.tar.gz")
    with gzip.open(compressed_container_path) as f:
        while True:
            chunk = f.read(chunk_size)
            if len(chunk) > 0:
                p.stdin.write(chunk)
            else:
                break
    p.communicate()

    if not is_container_installed():
        print("Failed to install the container image")
        return False

    print("Container image installed")
    return True
